@@ -1,2 +1,2 @@
 from palm_rlhf_pytorch.palm_rlhf_pytorch import PaLM, RewardModel, ActorWithValueHead
-from palm_rlhf_pytorch.ppg import RLHFTrainer
+from palm_rlhf_pytorch.ppo import RLHFTrainer
